import argparse
import code
import importlib.metadata
import itertools
import logging
import os
import re
import threading

import snisp

try:
    VERSION = importlib.metadata.version("SnakesInSpace")
except Exception:
    VERSION = ''


def regex_colors(arg):
    try:
        args = arg.split(', ')
        regex = re.compile(rf'{args[0]}')
        colors = itertools.cycle(args[1:])
        return regex, colors
    except Exception as e:
        raise argparse.ArgumentTypeError(
            f'{str(e)}: "regexes" expects an input of "regex,color", etc.'
        )


LOG_LEVELS = {
    'NOTSET': logging.NOTSET,
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
    'CRITICAL': logging.CRITICAL,
}

parser = argparse.ArgumentParser(
    prog='Snakes in Space',
    allow_abbrev=False,
    formatter_class=argparse.RawTextHelpFormatter,
    description='Snakes in Space implementation of SpaceTraders.',
    epilog='Additional information and basic HOW-TOs can be found here: '
           'https://github.com/PyWoody/SnakesInSpace'
           '\n\n'
           'For information regarding SpaceTraders, please visit their '
           'website at https://spacetraders.io/'
)
parser.add_argument(
    '-v', '-version', action='version', version=f'%(prog)s {VERSION}'
)
subparsers = parser.add_subparsers(
    title='subcommands',
    dest='subparsers',
    metavar='',
    help='See sub-command -h/--help for additional information'
)

interactive_parser = subparsers.add_parser(
    'interactive',
    allow_abbrev=False,
    formatter_class=argparse.RawTextHelpFormatter,
    help='Drops you into an interactive console with "agent" and '
         'command "ship" pre-loaded',
    description='This will drop you into an interactive shell of your current '
                'game. '
                'If you have already logged in at this location, '
                'you will be able to access your Agent as "agent". '
                'Otherwise, you will be dropped into an interactive '
                'session with only "snisp" imported.\n\n'
                'This is only intended for testing, status checking, etc. and '
                'not for actual playing.\n',
)
interactive_parser.add_argument(
    '-s',
    '--symbol',
    help='The symbol will act a the username. If left blank, \n'
         'the program will attempt to use the last logged in account. \n'
         'If left blank and no account has been logged in at this location \n'
         'before, a SpaceUserError will be raised.'
         '\n'
         '\n'
)
interactive_parser.add_argument(
    '-f',
    '--faction',
    default='COSMIC',
    choices=snisp.utils.FACTION_SYMBOL,
    metavar='',
    help='Default is "COSMIC". \n'
         'Using the non-standard Faction may cause issues. \n'
         'Using a different Faction than the one used to the '
         'account may cause other issues.'
         '\n'
         '\n'
         f'Options are: {", ".join(snisp.utils.FACTION_SYMBOL)}'
         '\n'
         '\n'
)
interactive_parser.add_argument(
    '-e',
    '--email',
    help='Optional'
)
interactive_parser.add_argument(
    '-t',
    '--token',
    help='Existing SpaceTraders token. Optional.'
)

tail_parser = subparsers.add_parser(
    'tail',
    allow_abbrev=False,
    formatter_class=argparse.RawTextHelpFormatter,
    help='Options for tailing the command log. '
         'By default, logs by "httpx" will be ignored.',
    description='Options for tailling the command log. '
                'By default, logs made by "httpx" will be ignored. '
                '\n\nIf you have "rich" (https://github.com/Textualize/rich) '
                'installed, you can add additonal regex patterns for '
                'highlighting.\nOtherwise, '
                'default highlights will be attempted.',
)
tail_parser.add_argument(
    '-p',
    '--patterns',
    nargs='+',
    type=regex_colors,
    help='Regex patterns to pass to Rich. Requires "rich" to be installed.\n'
         r'e.g.,--paterns  "\d+,blue on white,red on green" "\w,red,green"'
         '\nThe pattern is "{regex}, {colors},..." where {colors} will be'
         'cycled over all matches.'
)
tail_parser.add_argument(
    '-m',
    '--matches',
    nargs='+',
    help='List of regexes to filter the logs. Only logs that match any of '
         'the conditions in the regexes will be shown.'
)
tail_parser.add_argument(
    '-f',
    '--filters',
    nargs='+',
    help='List of regexes to filter the logs. Logs that match any of '
         'the conditions in the regexes will be NOT shown.'
)
tail_parser.add_argument(
    '--show-httpx',
    default=False,
    action='store_true',
    help='Flag to enable showing logs created by "httpx" in the tail output.'
)

logger_group = parser.add_argument_group('Logger Options')
no_logging_arg = logger_group.add_argument(
    '--no-logging',
    default=False,
    action='store_true',
    help='Disables the logger for this operation'
)
logger_group.add_argument(
    '--file-mode',
    choices=['w', 'a'],
    default='w',
    help='Sets the file mode for the logger to either "w" '
         'for write or "a" for append. Default is "w"'
)
logger_group.add_argument('--filename', help='Default is "snisp.log"')
logger_group.add_argument(
    '--level',
    choices=['NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
    metavar='NOTSET DEBUG INFO WARNING ERROR CRITICAL',
    default='INFO',
    help='Sets the logger level. Default is INFO'
)
logger_group.add_argument(
    '--exception-hook',
    choices=['thread_exception_hook_raiser', 'thread_exception_hook_logger'],
    metavar='thread_exception_hook_raiser thread_exception_hook_logger',
    default='thread_exception_hook_logger',
    help='The default, "thread_exception_hook_logger," will not raise '
         'exceptions that occur in threads\n'
         'but will merely log them as warnings.'
         '\n\n'
         '"thread_exception_hook_raiser" will print the error to terminal '
         'and attempt to raise it.'
)

args = parser.parse_args()

if args.filename:
    log_fname = args.filename
else:
    log_fname = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), 'snisp.log'
    )

if _subparser := args.subparsers:
    if args.exception_hook != 'thread_exception_hook_printer':
        threading.excepthook = snisp.utils.thread_exception_hook_raiser
    if args.no_logging:
        logging.getLogger().disabled = True
    else:
        logging.basicConfig(
            format='%(asctime)s %(levelname)s %(name)s | %(message)s',
            datefmt='%m/%d/%Y %I:%M:%S %p',
            encoding='utf-8',
            filename=log_fname,
            filemode=args.file_mode,
            level=LOG_LEVELS[args.level],
            force=True,
        )
    if _subparser == 'tail':
        if args.no_logging:
            raise argparse.ArgumentError(
                no_logging_arg,
                'Cannot run tail if there is nothing to tail. '
                'Please run the command again and remove "--no-logging"'
            )
        from snisp import tail
        tail.tail(
            log_fname,
            color_regexes=args.patterns,
            match_regexes=args.matches,
            filter_regexes=args.filters,
            show_httpx=args.show_httpx,
        )
    elif _subparser == 'interactive':
        try:
            agent = snisp.agent.Agent(
                symbol=args.symbol,
                faction=args.faction,
                email=args.email,
                token=args.token,
            )
            ship = next(iter(agent.fleet))
        except snisp.exceptions.SpaceUserError as e:
            print(f'WARNING: Could not Access Agent. {str(e)}')
        import readline  # noqa: F401
        code.InteractiveConsole(locals=locals()).interact()



import httpx
import inspect
import json
import os
import pytest

import snisp

from . import DATA_DIR


class TestSystems:

    agent = snisp.agent.Agent(symbol='testing', faction='testing')

    def test_repr(self, respx_mock):
        assert repr(self.agent.systems) == f'Systems({self.agent!r})'

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_list_systems(self, respx_mock):
        systems_data = json.load(
            open(os.path.join(DATA_DIR, 'systems.json'), encoding='utf8')
        )
        respx_mock.get('/systems', params={'page': 1}).mock(
            return_value=httpx.Response(200, json=systems_data)
        )
        respx_mock.get('/systems', params={'page': 2}).mock(
                return_value=httpx.Response(200, json={'data': []})
        )

        systems = list(self.agent.systems)
        assert len(systems) == 2
        assert systems[0].symbol == 'TEST-SYSTEM'
        assert systems[0].sectorSymbol == 'TEST'
        assert systems[1].factions[0].symbol == 'ANCIENTS'

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_find(self, respx_mock):
        systems_data = json.load(
            open(os.path.join(DATA_DIR, 'systems.json'), encoding='utf8')
        )
        respx_mock.get('/systems', params={'page': 1}).mock(
            return_value=httpx.Response(200, json=systems_data)
        )
        respx_mock.get('/systems', params={'page': 2}).mock(
                return_value=httpx.Response(200, json={'data': []})
        )

        system = self.agent.systems.find(type='NEUTRON_STAR')
        assert system.type == 'NEUTRON_STAR'
        assert system.symbol == 'TEST-SYSTEM'
        assert system.sectorSymbol == 'TEST'
        assert system.factions[0].symbol == 'COSMIC'

        with pytest.raises(snisp.exceptions.SpaceAttributeError):
            self.agent.systems.find(type='INVALID')

        assert not self.agent.systems.find(type='UNSTABLE')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_find_all(self, respx_mock):
        systems_data = json.load(
            open(os.path.join(DATA_DIR, 'systems.json'), encoding='utf8')
        )
        respx_mock.get('/systems', params={'page': 1}).mock(
            return_value=httpx.Response(200, json=systems_data)
        )
        respx_mock.get('/systems', params={'page': 2}).mock(
                return_value=httpx.Response(200, json={'data': []})
        )

        systems = list(self.agent.systems.find_all(type='BLACK_HOLE'))
        assert systems[0].type == 'BLACK_HOLE'
        assert systems[0].symbol == 'TEST-SYSTEM'
        assert systems[0].sectorSymbol == 'TEST'
        assert systems[0].factions[0].symbol == 'ANCIENTS'

        with pytest.raises(snisp.exceptions.SpaceAttributeError):
            next(self.agent.systems.find_all(type='INVALID'))


class TestSystem:

    agent = snisp.agent.Agent(symbol='testing', faction='testing')

    def test_repr(self, respx_mock):
        system = snisp.systems.System(self.agent)
        assert repr(system) == f'System({self.agent!r})'

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_call(self, respx_mock):
        system_data = json.load(
            open(os.path.join(DATA_DIR, 'system.json'), encoding='utf8')
        )
        respx_mock.get('/systems/TEST-SYSTEM').mock(
            return_value=httpx.Response(200, json=system_data)
        )

        system = snisp.systems.System(self.agent)
        system = system('TEST-SYSTEM')

        assert system.to_dict() == system_data['data']
        assert system.type == 'NEUTRON_STAR'
        assert system.symbol == 'TEST-SYSTEM'
        assert system.sectorSymbol == 'TEST'
        assert system.factions[0].symbol == 'COSMIC'

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_scan(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        scan_data = json.load(
            open(os.path.join(DATA_DIR, 'system_scan.json'), encoding='utf8')
        )
        respx_mock.post('/my/ships/TEST_SHIP_SYMBOL/scan/systems').mock(
            return_value=httpx.Response(200, json=scan_data)
        )
        orbit_data = json.load(
            open(os.path.join(DATA_DIR, 'orbit.json'), encoding='utf8')
        )
        respx_mock.post('/my/ships/TEST_SHIP_SYMBOL/orbit').mock(
            return_value=httpx.Response(200, json=orbit_data)
        )

        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        scans = ship.system.scan()
        assert inspect.isgenerator(scans)
        scans = list(scans)
        assert ship.cooldown.total_seconds == 60
        assert ship.cooldown.remaining_seconds == 59
        assert len(scans) == 1
        for scan, scan_json in zip(scans, scan_data['data']['systems']):
            assert scan.to_dict() == scan_json



import logging

from snisp import exceptions, fleet, utils, waypoints
from snisp.decorators import retry


logger = logging.getLogger(__name__)


class Shipyards:

    def __init__(self, ship, location):
        self.ship = ship
        self.agent = ship.agent
        self.location = location

    def __repr__(self):
        cls = self.__class__.__name__
        return f'{cls}({self.agent!r}, {self.location!r})'

    def __iter__(self):
        """
        Iterates over the Shipyard's in the Ship's current location

        Yields:
            Shipyard
        """
        waypoint = waypoints.Waypoints(self.agent, self.location)
        for shipyard in waypoint(traits='SHIPYARD'):
            yield Shipyard(self.agent, shipyard.to_dict())

    def autopurchase(self, *, ship_type, max_units=1, buffer=300_000):
        """
        Purchases up to max_units of ship_type, depending on
        the Agent's current credits with regards to the buffer and if the
        ship_type is currently available at a known Shipyard

        Buffer will be your credits buffer. The default 300_000 limit means you
        will be able to purchase up to max_units so long as your current
        agent.data.credits - buffer >= purchase price.
        To remove the buffer, just pass a 0

        NOTE: The only way to know the type and cost of a ship available
              for purchase at a Shipyard is to have a Ship or Probe located
              at the Shipyard. Just because nothing is returned, does not
              entail that no Ship's of that type are availble for purchase

        Args:
            ship_type: Ship type to purchase. See snisp.utils.SHIP_TYPES for
                       acceptable Ship types

        Kwargs:
            max_units: Up to max_units to purchase. Default is 1
            buffer: Credit buffer. Up to max_units will be purchased so far as
                    the purchase amount - agent.data.credits - buffer > 0.
                    Default is 300_000

        Returns:
            Transactions:  List of successful Transactions or an empty list
        """
        max_units = int(max_units)
        transactions = []
        smd = self.ship.markets.shipyard_market_data
        with self.agent.lock:
            while max_units > 0:
                shipyard, market_data = smd(ship_type)
                if not market_data or not shipyard:
                    logger.warning(
                        f'SYSTEM: {self.location.system} | '
                        f'No Shipyard found selling {ship_type}s'
                    )
                    return transactions
                credits_buffer = self.agent.data.credits - buffer
                if market_data.purchase_price <= credits_buffer:
                    transactions.append(shipyard.purchase(ship_type))
                    max_units -= 1
                else:
                    return transactions
        return transactions


class Shipyard(utils.AbstractJSONItem):

    def __init__(self, agent, shipyard):
        self.agent = agent
        self._data = shipyard

    @property
    def data(self):
        """Returns the Shipyard's ShipyardData"""
        response = self.agent.client.get(
            f'/systems/{self.system_symbol}/waypoints/'
            f'{self.symbol}/shipyard'
        )
        return ShipyardData(self.agent, response.json()['data'])

    @retry()
    def available_ships(self, ship_type=None):
        """
        Yields the ships available for purchasing at the Shipyard

        NOTE: The only way to know the type and cost of a ship available
              for purchase at a Shipyard is to have a Ship or Probe located
              at the Shipyard. Just because nothing is returned, does not
              entail that no Ship's of that type are availble for purchase
              at the Shipyard

        Kwargs:
            ship_type: Ship type to filter. See snisp.utils.SHIP_TYPES for
                       acceptable Ship types. Default is None
        Yields:
            ShipyardShip
        """
        if ship_type and ship_type not in utils.SHIP_TYPES:
            raise exceptions.SpaceAttributeError(
                f'{ship_type} is not an acceptable Ship type. '
                'See snisp.utils.SHIP_TYPES for acceptable types.'
            )
        response = self.agent.client.get(
            f'/systems/{self.system_symbol}'
            f'/waypoints/{self.symbol}/shipyard'
        )
        for ship in response.json()['data'].get('ships', []):
            if ship_type is not None and ship_type != ship['type']:
                continue
            if 'shipyard_symbol' not in ship.keys():
                ship['shipyard_symbol'] = self.symbol
            yield ShipyardShip(self.agent, ship)

    @retry()
    def purchase(self, ship_type):
        """
        Purchases a Ship of ship_type

        Args:
            ship_type: Ship type to filter. See snisp.utils.SHIP_TYPES for
                       acceptable Ship types
        Returns:
            Ship
        """
        if ship_type not in utils.SHIP_TYPES:
            raise exceptions.SpaceAttributeError(
                f'{ship_type} is not an acceptable Ship type. '
                'See snisp.utils.SHIP_TYPES for acceptable types.'
            )
        if not utils.a_ship_at_location(self.agent, self.symbol):
            cls = self.__class__.__name__
            raise exceptions.NoShipAtLocationError(
                f'No ships located at {self.symbol!r} for {cls}'
            )
        payload = {
            'shipType': ship_type, 'waypointSymbol': self.symbol
        }
        with self.agent.lock:
            response = self.agent.client.post('/my/ships', json=payload)
            data = response.json()['data']
            logger.info(
                f'Purchased a {data["transaction"]["shipType"]} at '
                f'{data["transaction"]["waypointSymbol"]} for '
                f'${data["transaction"]["price"]:,.2f}'
            )
            self.agent.recent_transactions.appendleft(
                fleet.Transaction(self.agent, data)
            )
            return fleet.Ship(self.agent, data['ship'])

    def transactions(self, page=1):
        """
        Yields the most recent Transactions at the Shipyard

        Requires a Ship or Probe to be located at the Shipyard

        Yields:
            Transaction
        """
        if not utils.a_ship_at_location(self.agent, self.symbol):
            cls = self.__class__.__name__
            raise exceptions.NoShipAtLocationError(
                f'No ships located at {self.symbol!r} for {cls}'
            )
        params = {'page': int(page)}
        response = self.agent.client.get(
            f'/systems/{self.system_symbol}/'
            f'waypoints/{self.symbol}/shipyard', params=params
        )
        transactions = response.json()['data'].get('transactions', [])
        while transactions:
            for transaction in transactions:
                yield fleet.Transaction(self.agent, transaction)
            params['page'] += 1
            response = self.agent.client.get(
                f'/systems/{self.system_symbol}/'
                f'waypoints/{self.symbol}/shipyard', params=params
            )
            transactions = response.json()['data'].get('transactions', [])


class ShipyardData(utils.AbstractJSONItem):

    def __init__(self, agent, data):
        self.agent = agent
        self._data = data


class ShipyardShip(utils.AbstractJSONItem):

    def __init__(self, agent, ship):
        self.agent = agent
        self._data = ship

    @retry()
    def purchase(self, *, buffer=200_000):
        """
        Purchases a Ship of the ShipyardShip's ship_type

        Convenience method used when iterating over available ship types
        for sale

        Buffer will be your credits buffer. The default 200_000 limit means you
        will be able to purchase a Ship so long as your current
        agent.data.credits - buffer >= purchase price.
        To remove the buffer, just pass a 0

        Kwargs:
            buffer: Credit buffer. Up to max_units will be purchased so far as
                    the purchase amount - agent.data.credits - buffer > 0.
                    Default is 200_000

        Returns:
            Ship
        """
        payload = {
            'shipType': self.type, 'waypointSymbol': self.shipyard_symbol
        }
        with self.agent.lock:
            if (self.agent.data.credits - buffer) > 0:
                response = self.agent.client.post('/my/ships', json=payload)
                data = response.json()['data']
                logger.info(
                    f'Purchased a {data["transaction"]["shipType"]} at '
                    f'{data["transaction"]["waypointSymbol"]} for '
                    f'${data["transaction"]["price"]:,.2f}'
                )
                return fleet.Ship(self.agent, data['ship'])
